#include <stdio.h>
#include <string.h>
#include <time.h>
#include "core.h"

#define ASCII1 "ascii1"
#define FILE_HEADER "Maxwell'sLemur"
#define BUF_HEADER "Buffer"
#define PAGE_HEADER "Page"
#define ROW_HEADER "Row"
#define CELL_HEADER "Cell"
#define FORMULA_HEADER "Formula"

static void buffer_load_line_ascii1(FILE *f, LoadStatus *ls) ;

gint buffer_save_ml(buffer *b, gchar *filename)
{
  gint res;
  FILE *f;
  
  f = fopen (filename, "w");
  
  res = write_header_ml(f, b);
  res = write_buffer_ml(f, b);
  fclose(f);
  return res;
}

gint write_header_ml(FILE *f, buffer *b) 
{
  fprintf(f, "%s:%s\n",FILE_HEADER,ASCII1);
  return 1;
}

gint write_buffer_ml(FILE *f, buffer *b)
{
  buffer_page *bp;
  fprintf(f, "%s:%s\n", BUF_HEADER, b->name);

  for (bp = b->pages; bp; bp = bp->next)
    {
      write_buffer_page_ml(f, b, bp);
    }
  return 1;
}

void write_buffer_page_ml(FILE *f, buffer *b, buffer_page *bp)
{
  gint lr, lc;
  gint row, col;
  fprintf(f, "%s:%s\n", PAGE_HEADER, bp->name);
  
  lr = row_last_used(bp);
  for (row = 1; row <= lr; row++) {
    if (bp->rows[row].nent_cells >0) {
      fprintf(f, "%s:%d\n", ROW_HEADER,row);
      lc = col_last_used(bp, row);
      for (col = 1; col <= lc; col++) {
	if (bp->rows[row].cells[col].ctype == CELL_TYPE_EXPRESSION) {
	  fprintf(f, "%s:%d\n", CELL_HEADER, col);
	  fprintf(f, "%s:%s\n", FORMULA_HEADER, bp->rows[row].cells[col].formula);
	}
      }
    }
  }      
}


buffer* buffer_load_ml(gchar *filename)
{
  FILE *f;
  LoadStatus ls;
  
  ls.b = NULL;
  ls.bp = NULL;
  ls.version = ls.error = 0;
  ls.filename = filename;

  f = fopen(filename, "r");
  while (!feof(f)) {
    buffer_load_line_ascii1(f, &ls);
  }
  fclose(f);
  calc_matrix(ls.b);
  return ls.b;
}

static void buffer_load_line_ascii1(FILE *f, LoadStatus *ls) 
{
  gchar *val;
  gchar tbuf[1000];
  fgets(tbuf,1000,f);
  val = strchr(tbuf,':');
  if (!val) return;
  *(val++) = 0;
  if (val[strlen(val)-1] == '\n') val[strlen(val)-1] = 0;
/*   fprintf(stderr, "%s -- %s\n", tbuf, val); */
  if (0 == strncmp(tbuf, FILE_HEADER, sizeof(FILE_HEADER))) {
    if (0 != strncmp(val, ASCII1, sizeof(ASCII1))) {
      fprintf (stderr, "Aeigh! The sky is falling!\n");
    }
  }
  if (0 == strncmp(tbuf, BUF_HEADER, sizeof(BUF_HEADER))) {
    if (!ls->b) {
      ls->b = buffer_new(buffer_name(val),ls->filename);
    } else {
      fprintf (stderr, "Eeeeh! The sky is falling!\n");
    }
  }
  if (0 == strncmp(tbuf, PAGE_HEADER, sizeof(PAGE_HEADER))) {
    ls->bp = bpage_new(ls->b, val);
  }
  if (0 == strncmp(tbuf, ROW_HEADER, sizeof(ROW_HEADER))) {
    ls->row = (gint) atoi(val);
  }
  if (0 == strncmp(tbuf, CELL_HEADER, sizeof(CELL_HEADER))) {
    ls->col = (gint) atoi(val);
  }
  if (0 == strncmp(tbuf, FORMULA_HEADER, sizeof(FORMULA_HEADER))) {
    ins_data(ls->bp, /* FIXME */ name2interpreter("Guile"),
	     val, CELL_TYPE_EXPRESSION,
	     0.0, VALUE_TYPE_UNCALCULATED,
	     ls->row, ls->col);
  }
  
}
