/*
  Maxwell's Lemur
  Copyright (c) 1997 Daniel Risacher <magnus@alum.mit.edu>

  Siag, Scheme In A Grid
  Copyright (C) 1996, 1997  Ulric Eriksson <ulric@edu.stockholm.se>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


/*
 * guilei.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include "core.h"

#include <guile/gh.h>

static int siag_row, siag_col;
static cval siag_result;
static buffer *siag_buffer;
static buffer_page *siag_bpage;

static int guile_interpreter;

static SCM get_row()
{
  return gh_int2scm(siag_row);
}

static SCM get_col()
{
  return gh_int2scm(siag_col);
}

static SCM x_set_cell(SCM formula, SCM row, SCM col, SCM bpage, SCM bname)
{
  gint r, c;
  buffer *b;
  buffer_page *bp;
  gchar *form;
 
  if (!gh_exact_p(row))
    scm_wta((long) row, "wta(1st) to get-cell", NULL);
  if (!gh_exact_p(col))
    scm_wta((long) col, "wta(2nd) to get-cell", NULL);
  r = (0.5 + gh_scm2double(row));
  c = (0.5 + gh_scm2double(col));
  if (SCM_NULLP(bname)) {
    b = siag_buffer;
  } else if (gh_string_p(bname)) {
    if (0 ==strlen(SCM_CHARS(bname))) 
      b = siag_buffer;
    else 
      b = find_buffer_by_name(SCM_CHARS(bname));
    if (b == NULL) return SCM_EOL;
  }
  else return SCM_EOL;

  if (SCM_NULLP(bpage)) {
    bp = b->pages;
  }
  else if (gh_string_p(bpage)) {
    if (0 ==strlen(SCM_CHARS(bpage))) 
      bp = b->pages;
    else 
      bp = find_bpage_by_name(b, SCM_CHARS(bpage));
    if (bp == NULL) return SCM_EOL;
  }
  else return SCM_EOL;
  form = SCM_CHARS(formula);
  
  ins_data(bp, guile_interpreter,
	   form, CELL_TYPE_EXPRESSION,
	   0.0, VALUE_TYPE_UNCALCULATED,
	   r,c);
  return formula;
}

static SCM x_get_cell(SCM row, SCM col, SCM bpage, SCM bname)
{
  gint r, c;
  gchar *p;
  buffer *b;
  buffer_page *bp;
  
  if (!gh_exact_p(row))
    scm_wta((long) row, "wta(1st) to get-cell", NULL);
  if (!gh_exact_p(col))
    scm_wta((long) col, "wta(2nd) to get-cell", NULL);
  r = (0.5 + gh_scm2double(row));
  c = (0.5 + gh_scm2double(col));
/*   if (r < 1 || r > BUFFER_ROWS || c < 1 || c > BUFFER_COLS) */
/*     return SCM_EOL; */
  if (SCM_NULLP(bname)) {
    b = siag_buffer;
  }
  else if (gh_string_p(bname)) {
    if (0 ==strlen(SCM_CHARS(bname))) 
      b = siag_buffer;
    else 
      b = find_buffer_by_name(SCM_CHARS(bname));
    if (b == NULL) return SCM_EOL;
  }
  else return SCM_EOL;

  if (SCM_NULLP(bpage)) {
    bp = siag_bpage;
  }
  else if (gh_string_p(bpage)) {
    if (0 ==strlen(SCM_CHARS(bpage))) 
      bp = siag_bpage;
    else 
      bp = find_bpage_by_name(b, SCM_CHARS(bpage));
    if (bp == NULL) return SCM_EOL;
  }
  else return SCM_EOL;
  
  switch (ret_vtype(bp, r, c)) {
  case VALUE_TYPE_STRING:
    p = ret_string(bp, r, c);
    return scm_makfrom0str( p);
  default:
    return gh_double2scm(ret_val(bp, r, c).number);
  }
}

static SCM get_cell(SCM row, SCM col, SCM bname, SCM bpage)
{
  if (SCM_UNBNDP(bname))
    bname = SCM_EOL;
  if (SCM_UNBNDP(bpage))
    bpage = SCM_EOL;
  return x_get_cell(row, col, bname, bpage);
}

static SCM x_get_string(SCM row, SCM col, SCM bname, SCM bpage)
{
  int r, c;
  char *p;
  buffer *b;
  buffer_page *bp;
  
  if (!gh_exact_p(row))
    scm_wta((long) row, "wta(1st) to get-string", NULL);
  if (!gh_exact_p(col))
    scm_wta((long) col, "wta(2nd) to get-string", NULL);
  r = (0.5+gh_scm2double(row));
  c = (0.5+gh_scm2double(col));
/*   if (r < 1 || r > BUFFER_ROWS || c < 1 || c > BUFFER_COLS) */
/*     return SCM_EOL; */

  if (SCM_NULLP(bname))
    b = siag_buffer;
  else if (gh_string_p(bname)) {
    b = find_buffer_by_name(SCM_CHARS(bname));
    if (b == NULL) return SCM_EOL;
  }
  else return SCM_EOL;
  
  if (SCM_NULLP(bpage)) {
    bp = b->pages;
  }
  else if (gh_string_p(bpage)) {
    bp = find_bpage_by_name(b, SCM_CHARS(bpage));
    if (bp == NULL) return SCM_EOL;
  }
  else return SCM_EOL;
  
  if (ret_vtype(bp, r, c) == VALUE_TYPE_ERROR) p = "";
  else p = ret_pvalue(NULL, bp, r, c, -1);
  return scm_makfrom0str(p);
}

/* FIXME */
static SCM get_string(SCM row, SCM col)
{
  return x_get_string(row, col, SCM_EOL, SCM_EOL);
}

/* FIXME: translate constants to symbols */
static SCM get_ctype(SCM bname, SCM row, SCM col)
{
  return gh_double2scm(ret_ctype(siag_bpage, gh_scm2double(row), gh_scm2double(col)));
}

static SCM get_vtype(SCM bname, SCM row, SCM col)
{
  return gh_double2scm(ret_vtype(siag_bpage, gh_scm2double(row), gh_scm2double(col)));
}

static SCM get_text(SCM row, SCM col)
{
  char *text;

  text = ret_taddr(siag_bpage, gh_scm2double(row), gh_scm2double(col));
  if (!text) text = "";
  
  return scm_makfrom0str( text);
}

static SCM siag_time()
{
	double t = time(NULL);
	return gh_double2scm(t);
}

static SCM lexec_expr(SCM intp, SCM expr)
{
	exec_expr(name2interpreter(SCM_CHARS(intp)), SCM_CHARS(expr));
	return SCM_EOL;
}

#if 0
static void siag_puts(char *p)
{
}
#endif

static void siag_print(SCM p)
{
  if (gh_number_p(p)) {
    siag_type = VALUE_TYPE_NUMBER;
    siag_result.number = gh_scm2double(p);
  } else if (gh_string_p(p)) {
    siag_type = VALUE_TYPE_STRING;
    siag_result.text = SCM_CHARS(p);
  } else {
    siag_type = VALUE_TYPE_ERROR;
    siag_result.number = 0, errorflag = 1;
  }
}

/* Using Guile */

#define BREAKCHARS "() \t\r\n"

static char *update_Guile_references(char *expr, int r1, int c1,
				     int r2, int c2, int rd, int cd)
{
  return expr;
}


static char *expand_references(char *orig)
{
  char *new;
  int rc = ref_counter(orig);
  if (!rc) return orig;
  
  /* strlen("(get-cell 1 1 \"page\" \"buffer\")")-strlen("buffer:page:r1c1") = 14 */
  new = g_malloc(strlen(orig)+14*rc+1);
  ref_expander(orig, new, BREAKCHARS, "(get-cell %ld %ld \"%s\" \"%s\")");
  return new;
}

#if 0

/* move block between (r1,c1) and (r2,c2), direction (rd,cd) */
static char *update_Guile_references(char *expr, int r1, int c1,
				     int r2, int c2, int rd, int cd)
{
  char *p;
  int rc = ref_counter(expr);
  
  if (!rc) return expr;
  
  /* rough upper bound on new length. A few bytes extra don't matter */
  /* strlen("r100000c100000")-strlen("r1c1") = 10 */
  p = malloc(strlen(expr)+10*rc+1);
  rc = ref_updater(expr, p, BREAKCHARS, r1, c1, r2, c2, rd, cd);
  if (rc) {
    expr = strdup(p);
  }
  free(p);
  return expr;
}
#endif 

static SCM wrapper(void *data, SCM jmpbuf)
{
  char *scheme_code = (char *)data;
  char *expansion = expand_references(scheme_code); 
/*   char *expansion = scheme_code;  */
  
  SCM res = gh_eval_str(expansion);
  if (expansion != scheme_code) free (expansion);
  return res;
}

static SCM catcher(void *data, SCM tag, SCM throw_args)
{
  char b[256];
  
  strcpy(b, "ERROR: ");
  if (data) 
    strncat(b, (char *)data, 200);
  /* 	llpr(b); */
  return SCM_BOOL_F;
}

/* This same function also does strings */
cval parse_guile_expr(buffer *b, buffer_page *bp, char *expr, gint row, gint col)
{
  SCM result;
  
  siag_row = row;
  siag_col = col;
  siag_buffer = b;
  siag_bpage = bp;

  /* the next two lines seem to contradict each other, but it is
     necessary to set the type in case siod bails out before
     we make it to siag_print(). */
  errorflag = 0;
  siag_type = VALUE_TYPE_ERROR;

  result = gh_catch(SCM_BOOL_T,
		    (scm_catch_body_t)wrapper, expr,
		    (scm_catch_handler_t)catcher, expr);
  siag_print(result);
  return siag_result;
}

void execute_guile(char *s, gint row, gint col, buffer *b, buffer_page *bp)
{
  siag_row = row;
  siag_col = col;
  siag_buffer = b;
  siag_bpage = bp;
  gh_catch(SCM_BOOL_T,
	   (scm_catch_body_t)wrapper, s,
	   (scm_catch_handler_t)catcher, s);
}

gint init_guile_parser()
{
  gh_new_procedure0_0("row", get_row);
  gh_new_procedure0_0("col", get_col);
  gh_new_procedure("get-cell", get_cell,2,2,0);
  gh_new_procedure2_0("get-string", get_string);
  gh_new_procedure3_0("get-ctype", get_ctype);
  gh_new_procedure3_0("get-vtype", get_vtype);
  gh_new_procedure2_0("get-text", get_text);
/*   gh_new_procedure2_0("sum", siag_sum); */
  gh_new_procedure4_0("x-get-cell", x_get_cell);
  gh_new_procedure("x-set-cell", x_set_cell, 5,0,0);
  gh_new_procedure4_0("x-get-string", x_get_string);
  gh_new_procedure0_0("time", siag_time);
  gh_new_procedure2_0("exec-expr", lexec_expr);
  
  return guile_interpreter = register_interpreter("Guile",
						  parse_guile_expr, 
						  execute_guile,
						  update_Guile_references);
}


